from re import Pattern
from typing import Type, TypeVar

from partial_json_parser import ALL, JSON, Allow

F = TypeVar("F")
M = TypeVar("M")

def extract_json(
    text: str,
    fallback: F = None,
    expect: Type[M] = JSON,
    allow_partial: Allow = ALL,
) -> M | F: ...

json_block_pattern: Pattern = ...

def find_json_blocks(text: str) -> list[str]: ...



from typing import Literal, NotRequired, TypedDict, overload

from tiktoken import Encoding

class Message(TypedDict):
    role: Literal["system", "user", "assistant"]
    content: str
    name: NotRequired[str]

def cached_get_encoding(encoding_name: str) -> Encoding: ...
@overload
def count_token(prompt: str, enc: Encoding | None = None) -> int: ...
@overload
def count_token(prompt: Message, enc: Encoding | None = None) -> int: ...
@overload
def count_token(prompt: list[str], enc: Encoding | None = None) -> int: ...
@overload
def count_token(prompt: list[Message], enc: Encoding | None = None) -> int: ...



"""Count tokens for OpenAI models"""

from functools import lru_cache

from tiktoken import get_encoding

cached_get_encoding = lru_cache(1)(get_encoding)


def count_token(prompt, enc=None):
    """count number of tokens in prompt"""

    enc = enc or cached_get_encoding("cl100k_base")

    if isinstance(prompt, str):
        return len(enc.encode(prompt, disallowed_special=()))

    if isinstance(prompt, list):
        if len(prompt) == 0:
            return 0
        return sum(count_token(i, enc) for i in prompt) + (0 if isinstance(prompt[0], str) else 3)

    if isinstance(prompt, dict):
        return 3 + (len(enc.encode(f"name={prompt['name']}")) if "name" in prompt else 1) + len(enc.encode(prompt["content"]))

    raise TypeError(prompt)



from contextlib import suppress
from re import DOTALL, compile

from partial_json_parser import ALL, loads


def extract_json(text: str, fallback=None, expect=None, allow_partial=ALL):
    """parse JSON from raw LLM response text"""

    with suppress(IndexError):
        text = find_json_blocks(text)[-1]  # choose the last one

    with suppress(ValueError):
        result = loads(text, allow_partial)
        if expect is None:
            return result

        from pydantic import BaseModel, TypeAdapter, ValidationError

        with suppress(ValidationError):
            with suppress(TypeError):
                if issubclass(expect, BaseModel):
                    return expect.model_validate(result)
            return TypeAdapter(expect).validate_python(result)

    return fallback


json_block_pattern = compile(r"```json\n(.+?)(?:\n```|`{0,3}\Z)", DOTALL)


def find_json_blocks(text):
    """find json code blocks from text"""
    return json_block_pattern.findall(text)



from itertools import count
from typing import Any, AsyncIterator, Callable, Iterator, ParamSpec, TypeVar, overload

from attrs import define
from promplate import Template

_template: str

global_count: count

event_template: Template

@define(slots=True, hash=True)
class Event:
    data: str
    event: Any = None
    id: Callable[[], Any] | Any = global_count.__next__

AnyIterator = TypeVar("AnyIterator", Iterator[str], AsyncIterator[str])
P = ParamSpec("P")

type SyncAnyIterator = Iterator[str] | Iterator[tuple[str, Any]] | Iterator[tuple[str, Any, Callable[[], Any] | Any]]

type AsyncAnyIterator = AsyncIterator[str] | AsyncIterator[tuple[str, Any]] | AsyncIterator[tuple[str, Any, Callable[[], Any] | Any]]

@overload
def as_event_stream(gen: SyncAnyIterator) -> Iterator[Event]: ...
@overload
def as_event_stream(gen: AsyncAnyIterator) -> AsyncIterator[Event]: ...
@overload
def as_event_stream(gen: Callable[P, SyncAnyIterator]) -> Callable[P, Iterator[Event]]: ...
@overload
def as_event_stream(gen: Callable[P, AsyncAnyIterator]) -> Callable[P, AsyncIterator[Event]]: ...



from functools import wraps
from inspect import isasyncgen, isasyncgenfunction, isgenerator
from itertools import count
from typing import Any, Callable, Union

from attrs import asdict, define
from promplate import Template

_template = """\
{% if callable(id) -%}
id: {{ id() }}
{% elif id is not None -%}
id: {{ id }}
{% endif -%}
{% if event is not None -%}
event: {{ event }}
{% endif -%}
{% for line in data.splitlines() -%}
data: {{ line }}
{% endfor %}
"""

global_count = count()

event_template = Template(_template)


@define(slots=True, hash=True)
class Event:
    data: str
    event: Any = None
    id: Union[Callable[[], Any], Any] = global_count.__next__

    def __str__(self):
        return event_template.render(asdict(self))


def as_event_stream(gen):
    if isgenerator(gen):

        def _():
            for i in gen:
                yield Event(*i) if isinstance(i, tuple) else Event(i)

        return _()

    if isasyncgen(gen):

        async def _():
            async for i in gen:
                yield Event(*i) if isinstance(i, tuple) else Event(i)

        return _()

    assert callable(gen)

    if isasyncgenfunction(gen):

        @wraps(gen)
        async def _(*args, **kwargs):
            async for i in gen(*args, **kwargs):
                yield Event(*i) if isinstance(i, tuple) else Event(i)

        return _

    @wraps(gen)
    def _(*args, **kwargs):
        return as_event_stream(gen(*args, **kwargs))

    return _
